import {
  EMPTY_STRING,
  LinkRelationName,
  ResourceRepository,
  ResourceServiceConfig,
  StateResource,
  createStateResource,
} from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { Resource } from '@ngxp/rest';
import { Observable, of } from 'rxjs';
import { createCommandResource } from '../../../command-shared/test/command';
import { singleCold, singleHot } from '../../../tech-shared/test/marbles';
import { createDummyResource } from '../../../tech-shared/test/resource';
import { CommandResourceService } from './command-resource.service';
import { CommandResource } from './command.model';
import { CommandService } from './command.service';

describe('CommandResourceService', () => {
  let service: CommandResourceService<Resource, Resource>;
  let config: ResourceServiceConfig<Resource>;
  let repository: Mock<ResourceRepository>;
  let commandService: Mock<CommandService>;

  const configResource: Resource = createDummyResource();
  const configStateResource: StateResource<Resource> = createStateResource(configResource);
  const configStateResource$: Observable<StateResource<Resource>> = of(configStateResource);

  const editLinkRel: string = 'dummyEditLinkRel';
  const getLinkRel: LinkRelationName = 'dummyGetLinkRel';

  const deleteOrder: string = 'dummyDeleteOrder';
  const deleteLinkRel: string = 'dummyDeleteLinkRel';

  beforeEach(() => {
    config = {
      resource: configStateResource$,
      getLinkRel,
      edit: { linkRel: editLinkRel },
      delete: { order: deleteOrder, linkRel: deleteLinkRel },
    };
    repository = mock(ResourceRepository);
    commandService = mock(CommandService);

    service = new CommandResourceService(
      config,
      useFromMock(repository),
      useFromMock(commandService),
    );
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('delete', () => {
    const resourceWithDeleteLinkRel: Resource = createDummyResource([deleteLinkRel]);
    const stateResourceWithDeleteLink: StateResource<Resource> =
      createStateResource(resourceWithDeleteLinkRel);

    beforeEach(() => {
      commandService.createCommandByProps.mockReturnValue(
        of(createStateResource(createCommandResource())),
      );
      service.stateResource.next(stateResourceWithDeleteLink);
    });

    it('should throw error if delete linkRel not exists on current stateresource', () => {
      service.stateResource.next(createStateResource(createDummyResource()));

      expect(() => service.delete()).toThrowError(
        'No delete link exists on current stateresource.',
      );
    });

    it('should call command service', () => {
      service.delete();

      expect(commandService.createCommandByProps).toHaveBeenCalledWith({
        resource: resourceWithDeleteLinkRel,
        linkRel: deleteLinkRel,
        command: { order: deleteOrder, body: null },
        snackBarMessage: EMPTY_STRING,
      });
    });

    it('should return value', () => {
      const deleteResource: Resource = createDummyResource();
      const deleteStateResource: StateResource<Resource> = createStateResource(deleteResource);
      commandService.createCommandByProps.mockReturnValue(singleHot(deleteStateResource));

      const deletedResource: Observable<StateResource<CommandResource>> = service.delete();

      expect(deletedResource).toBeObservable(singleCold(deleteStateResource));
    });
  });
});
