import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { HttpErrorResponse } from '@angular/common/http';
import { fakeAsync, tick } from '@angular/core/testing';
import { faker } from '@faker-js/faker';
import { Resource, ResourceUri, getUrl } from '@ngxp/rest';
import { cold } from 'jest-marbles';
import { Observable, lastValueFrom, of, throwError } from 'rxjs';
import { createProblemDetail } from '../../../test//error';
import { singleCold, singleHot } from '../../../test/marbles';
import { createDummyResource } from '../../../test/resource';
import { HttpError, ProblemDetail } from '../tech.model';
import { LinkRelationName, ResourceServiceConfig } from './resource.model';
import { ResourceRepository } from './resource.repository';
import {
  StateResource,
  createEmptyStateResource,
  createErrorStateResource,
  createStateResource,
} from './resource.util';

import { ResourceService } from './resource.service';
import * as ResourceUtil from './resource.util';

describe('ResourceService', () => {
  let service: DummyResourceService<Resource, Resource>;
  let config: ResourceServiceConfig<Resource>;

  let repository: Mock<ResourceRepository>;

  const configResource: Resource = createDummyResource();
  const configStateResource: StateResource<Resource> = createStateResource(configResource);
  const configStateResource$: Observable<StateResource<Resource>> = of(configStateResource);

  const editLinkRel: string = 'dummyEditLinkRel';
  const getLinkRel: LinkRelationName = 'dummyGetLinkRel';
  const deleteLinkRel: LinkRelationName = 'dummyDeleteLinkRel';

  beforeEach(() => {
    config = {
      resource: configStateResource$,
      getLinkRel,
      edit: { linkRel: editLinkRel },
      delete: { linkRel: deleteLinkRel },
    };
    repository = mock(ResourceRepository);

    service = new DummyResourceService(config, useFromMock(repository));
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('get', () => {
    const stateResource: StateResource<Resource> = createStateResource(configResource);
    let isInvalidResourceCombinationSpy: jest.SpyInstance;

    beforeEach(() => {
      service.stateResource.next(stateResource);

      service.handleNullConfigResource = jest.fn();
      service.handleConfigResourceChanged = jest.fn();
      isInvalidResourceCombinationSpy = jest
        .spyOn(ResourceUtil, 'isInvalidResourceCombination')
        .mockReturnValue(true);
    });

    it('should handle config resource changed', fakeAsync(() => {
      service.get().subscribe();
      tick();

      expect(service.handleConfigResourceChanged).toHaveBeenCalledWith(
        stateResource,
        configResource,
      );
    }));

    it('should handle null configresource', fakeAsync(() => {
      service.get().subscribe();
      tick();

      expect(service.handleNullConfigResource).toHaveBeenCalledWith(configResource);
    }));

    it('should call isInvalidResourceCombinationSpy', fakeAsync(() => {
      service.get().subscribe();
      tick();

      expect(isInvalidResourceCombinationSpy).toHaveBeenCalled();
    }));

    it('should return initial value', () => {
      service.stateResource.asObservable = jest
        .fn()
        .mockReturnValue(singleHot(stateResource, '-a'));

      const apiRootStateResource$: Observable<StateResource<Resource>> = service.get();

      expect(apiRootStateResource$).toBeObservable(
        cold('a', { a: createEmptyStateResource(true) }),
      );
    });
  });

  describe('handle config resource changed', () => {
    const stateResource: StateResource<Resource> = createStateResource(createDummyResource());
    const changedConfigResource: Resource = createDummyResource();

    describe('on different config resource', () => {
      it('should update config resource if is different', () => {
        service.configResource = createDummyResource();

        service.handleConfigResourceChanged(stateResource, changedConfigResource);

        expect(service.configResource).toBe(changedConfigResource);
      });

      it('should set state resource reload', () => {
        service.configResource = createDummyResource();

        service.handleConfigResourceChanged(stateResource, changedConfigResource);

        expect(service.stateResource.value.reload).toBeTruthy();
      });
    });

    describe('on same config resource', () => {
      const stateResource: StateResource<Resource> = createStateResource(createDummyResource());

      beforeEach(() => {
        service.configResource = configResource;
      });

      it('should call shouldLoadResource', () => {
        service.shouldLoadResource = jest.fn();

        service.handleConfigResourceChanged(stateResource, configResource);

        expect(service.shouldLoadResource).toHaveBeenCalledWith(stateResource, configResource);
      });

      it('should load resource', () => {
        service.shouldLoadResource = jest.fn().mockReturnValue(true);
        service.loadResource = jest.fn();

        service.handleConfigResourceChanged(stateResource, configResource);

        expect(service.loadResource).toHaveBeenCalledWith(configResource);
      });

      it('should NOT load resource on shouldLoadResource false', () => {
        service.loadResource = jest.fn();
        service.shouldLoadResource = jest.fn().mockReturnValue(false);

        service.handleConfigResourceChanged(stateResource, configResource);

        expect(service.loadResource).not.toHaveBeenCalled();
      });
    });
  });

  describe('handle null config resource', () => {
    const resource: Resource = createDummyResource();
    const stateResource: StateResource<Resource> = createStateResource(resource);

    beforeEach(() => {
      service.shouldClearStateResource = jest.fn();
      service.stateResource.next(stateResource);
    });
    it('should call shouldClearStateResource', () => {
      service.handleNullConfigResource(null);

      expect(service.shouldClearStateResource).toHaveBeenCalledWith(null);
    });

    it('should clear stateresource if shouldClearStateResource is true', () => {
      service.shouldClearStateResource = jest.fn().mockReturnValue(true);

      service.handleNullConfigResource(null);

      expect(service.stateResource.value).toEqual(createEmptyStateResource());
    });

    it('should keep stateresource if shouldClearStateResource is false', () => {
      service.shouldClearStateResource = jest.fn().mockReturnValue(false);

      service.handleNullConfigResource(null);

      expect(service.stateResource.value).toBe(stateResource);
    });
  });

  describe('should clear stateresource', () => {
    const resource: Resource = createDummyResource();
    const stateResource: StateResource<Resource> = createStateResource(resource);

    it('should return true on null configresource and filled stateresource', () => {
      service.stateResource.next(stateResource);

      const shouldClear: boolean = service.shouldClearStateResource(null);

      expect(shouldClear).toBeTruthy();
    });

    it('should return false on null configresource and empty stateresource', () => {
      service.stateResource.next(createEmptyStateResource());

      const shouldClear: boolean = service.shouldClearStateResource(null);

      expect(shouldClear).toBeFalsy();
    });
  });

  describe('should load resource', () => {
    const resource: Resource = createDummyResource();
    const stateResource: StateResource<Resource> = createStateResource(resource);

    let isLoadingRequiredSpy: jest.SpyInstance<boolean>;

    beforeEach(() => {
      isLoadingRequiredSpy = jest.spyOn(ResourceUtil, 'isLoadingRequired');
    });

    it('should return true on existing configresource and loading is required', () => {
      isLoadingRequiredSpy.mockReturnValue(true);

      const shouldLoad: boolean = service.shouldLoadResource(stateResource, configResource);

      expect(shouldLoad).toBeTruthy();
    });

    it('should call isLoadingRequired', () => {
      service.shouldLoadResource(stateResource, configResource);

      expect(isLoadingRequiredSpy).toBeCalledWith(stateResource);
    });

    it('should return false if configresource exists but loading is NOT required', () => {
      isLoadingRequiredSpy.mockReturnValue(false);

      const shouldLoad: boolean = service.shouldLoadResource(stateResource, configResource);

      expect(shouldLoad).toBeFalsy();
    });
  });

  describe('load resource', () => {
    const configResourceWithGetLinkRel: Resource = createDummyResource([getLinkRel]);

    it('should throw error if getLinkRel not exists', () => {
      expect(() => service.loadResource(configResource)).toThrowError(
        'No get link exists on configresource.',
      );
    });

    it('should call do load resource', () => {
      service.doLoadResource = jest.fn();

      service.loadResource(configResourceWithGetLinkRel);

      expect(service.doLoadResource).toHaveBeenCalledWith(
        getUrl(configResourceWithGetLinkRel, config.getLinkRel),
      );
    });
  });

  describe('set state resource loading', () => {
    it('should set loading true', () => {
      service.stateResource.next(createStateResource(createDummyResource()));

      service.setStateResourceLoading();

      expect(service.stateResource.value.loading).toBeTruthy();
    });

    it('should set reload false', () => {
      service.stateResource.next({ ...createStateResource(createDummyResource()), reload: true });

      service.setStateResourceLoading();

      expect(service.stateResource.value.reload).toBeFalsy();
    });
  });

  describe('do load resource', () => {
    let resourceUri: ResourceUri;
    let loadedResource: Resource;

    beforeEach(() => {
      service.setStateResourceLoading = jest.fn();
      resourceUri = faker.internet.url();
      loadedResource = createDummyResource();
      repository.getResource.mockReturnValue(of(loadedResource));
    });

    it('should set state resource to loading', () => {
      service.doLoadResource(resourceUri);

      expect(service.setStateResourceLoading).toHaveBeenCalled();
    });

    it('should call repository', () => {
      service.doLoadResource(resourceUri);

      expect(repository.getResource).toHaveBeenCalledWith(resourceUri);
    });

    it('should update stateresource', () => {
      service.updateStateResource = jest.fn();

      service.doLoadResource(resourceUri);

      expect(service.updateStateResource).toHaveBeenCalledWith(loadedResource);
    });
  });

  describe('setResourceByUri', () => {
    it('should do load resource', () => {
      service.doLoadResource = jest.fn();
      const resourceUri: ResourceUri = faker.internet.url();

      service.setResourceByUri(resourceUri);

      expect(service.doLoadResource).toHaveBeenCalledWith(resourceUri);
    });
  });

  describe('update stateresource', () => {
    const resourceToBeSet: Resource = createDummyResource();

    it('should set resource as stateresource', () => {
      service.stateResource.next(createEmptyStateResource());

      service.updateStateResource(resourceToBeSet);

      expect(service.stateResource.value).toEqual(createStateResource(resourceToBeSet));
    });
  });

  describe('save', () => {
    const dummyToSave: unknown = {};
    const loadedResource: Resource = createDummyResource();

    const resourceWithEditLinkRel: Resource = createDummyResource([editLinkRel]);

    it('should throw error if edit link not exists', () => {
      service.stateResource.next(createStateResource(createDummyResource()));

      expect(() => service.save(dummyToSave)).toThrowError(
        'No edit link exists on current stateresource.',
      );
    });

    it('should do save', fakeAsync(() => {
      const stateResource: StateResource<Resource> = createStateResource(resourceWithEditLinkRel);
      service.stateResource.next(stateResource);
      const doSaveMock: jest.Mock = (service.doSave = jest.fn()).mockReturnValue(
        of(loadedResource),
      );

      service.save(dummyToSave).subscribe();
      tick();

      expect(doSaveMock).toHaveBeenCalledWith(resourceWithEditLinkRel, dummyToSave);
    }));

    it('should return saved object', () => {
      service.stateResource.next(createStateResource(resourceWithEditLinkRel));
      service.doSave = jest.fn().mockReturnValue(singleHot(loadedResource));

      const saved: Observable<StateResource<Resource | HttpError>> = service.save(dummyToSave);

      expect(saved).toBeObservable(singleCold(createStateResource(loadedResource)));
    });

    it('should call handleError', () => {
      service.stateResource.next(createStateResource(createDummyResource([config.edit.linkRel])));
      const errorResponse: ProblemDetail = createProblemDetail();
      service.doSave = jest.fn().mockReturnValue(throwError(() => errorResponse));
      service.handleError = jest.fn();

      service.save(<any>{}).subscribe();

      expect(service.handleError).toHaveBeenCalledWith(errorResponse);
    });

    it('should update state resource subject', fakeAsync(() => {
      service.stateResource.next(createStateResource(resourceWithEditLinkRel));
      service.doSave = jest.fn().mockReturnValue(of(loadedResource));

      service.save(dummyToSave).subscribe();
      tick();

      expect(service.stateResource.value).toEqual(createStateResource(loadedResource));
    }));
  });

  describe('handleError', () => {
    it('should return error stateresource on problem unprocessable entity', (done: jest.DoneCallback) => {
      const error: ProblemDetail = createProblemDetail();

      service
        .handleError(<HttpErrorResponse>(<any>error))
        .subscribe((responseError: StateResource<HttpError>) => {
          expect(responseError).toEqual(createErrorStateResource(error));
          done();
        });
    });

    it('should rethrow error', () => {
      const error: HttpErrorResponse = <HttpErrorResponse>{
        status: 500,
        statusText: 'Internal Server Error',
      };

      const thrownError$: Observable<StateResource<HttpError>> = service.handleError(error);

      expect.assertions(1);
      expect(lastValueFrom(thrownError$)).rejects.toThrowError('Internal Server Error');
    });
  });

  describe('refresh', () => {
    beforeEach(() => {
      service.loadResource = jest.fn();
    });

    it('should set reload true on statresource', () => {
      service.stateResource.next(createStateResource(createDummyResource()));

      service.refresh();

      expect(service.stateResource.value.reload).toBeTruthy();
    });
  });

  describe('can edit', () => {
    it('should return true if link is present', () => {
      const resource: StateResource<Resource> = createStateResource(
        createDummyResource([editLinkRel]),
      );
      service.stateResource.next(resource);

      const canEdit: boolean = service.canEdit();

      expect(canEdit).toBeTruthy();
    });

    it('should return false if link is NOT present', () => {
      const resource: StateResource<Resource> = createStateResource(createDummyResource());
      service.stateResource.next(resource);

      const canEdit: boolean = service.canEdit();

      expect(canEdit).toBeFalsy();
    });
  });

  describe('can delete', () => {
    it('should return true if link is present', () => {
      const resource: StateResource<Resource> = createStateResource(
        createDummyResource([deleteLinkRel]),
      );
      service.stateResource.next(resource);

      const canEdit: boolean = service.canDelete();

      expect(canEdit).toBeTruthy();
    });

    it('should return false if link is NOT present', () => {
      const resource: StateResource<Resource> = createStateResource(createDummyResource());
      service.stateResource.next(resource);

      const canEdit: boolean = service.canDelete();

      expect(canEdit).toBeFalsy();
    });
  });

  describe('get resource', () => {
    it('should return resource from stateResource', () => {
      const resource: Resource = createDummyResource();
      const stateResource: StateResource<Resource> = createStateResource(resource);
      service.stateResource.next(stateResource);

      const result: Resource = service.getResource();

      expect(result).toBe(resource);
    });
  });

  describe('exists', () => {
    it('should return true', () => {
      service.updateStateResource(createDummyResource());

      const exists = service.exists();

      expect(exists).toBeTruthy();
    });

    it('should return false', () => {
      service.updateStateResource(null);

      const exists = service.exists();

      expect(exists).toBeFalsy();
    });
  });
});

export class DummyResourceService<B extends Resource, T extends Resource> extends ResourceService<
  B,
  T
> {
  constructor(
    protected config: ResourceServiceConfig<B>,
    protected repository: ResourceRepository,
  ) {
    super(config, repository);
  }

  doSave(resource: T, toSave: unknown): Observable<T> {
    return of(resource);
  }
}
