import { HttpErrorResponse } from '@angular/common/http';
import { getUrl, hasLink, Resource, ResourceUri } from '@ngxp/rest';
import { isEqual, isNull } from 'lodash-es';
import {
  BehaviorSubject,
  catchError,
  combineLatest,
  filter,
  first,
  map,
  Observable,
  of,
  startWith,
  tap,
  throwError,
} from 'rxjs';
import { isUnprocessableEntity } from '../http.util';
import { HttpError } from '../tech.model';
import { isNotNull } from '../tech.util';
import { ResourceServiceConfig } from './resource.model';
import { ResourceRepository } from './resource.repository';
import { mapToFirst, mapToResource } from './resource.rxjs.operator';
import {
  createEmptyStateResource,
  createErrorStateResource,
  createStateResource,
  isInvalidResourceCombination,
  isLoadingRequired,
  StateResource,
  throwErrorOn,
} from './resource.util';

/**
 * B = Type of baseresource
 * T = Type of the resource which is working on
 */
export abstract class ResourceService<B extends Resource, T extends Resource> {
  readonly stateResource: BehaviorSubject<StateResource<T>> = new BehaviorSubject(
    createEmptyStateResource(),
  );

  configResource: B;

  constructor(
    protected config: ResourceServiceConfig<B>,
    protected repository: ResourceRepository,
  ) {}

  public get(): Observable<StateResource<T>> {
    return combineLatest([this.stateResource.asObservable(), this.getConfigResource()]).pipe(
      tap(([stateResource, configResource]) =>
        this.handleConfigResourceChanged(stateResource, configResource),
      ),
      tap(([, configResource]) => this.handleNullConfigResource(configResource)),
      filter(
        ([stateResource]) => !isInvalidResourceCombination(stateResource, this.configResource),
      ),
      mapToFirst<T, B>(),
      startWith(createEmptyStateResource<T>(true)),
    );
  }

  private getConfigResource(): Observable<B> {
    return this.config.resource.pipe(
      filter(
        (configStateResource: StateResource<B>) =>
          !configStateResource.loading && !configStateResource.reload,
      ),
      mapToResource<B>(),
    );
  }

  handleConfigResourceChanged(stateResource: StateResource<T>, configResource: B): void {
    if (!isEqual(this.configResource, configResource)) {
      this.configResource = configResource;
      this.stateResource.next({ ...this.stateResource.value, reload: true });
    } else if (this.shouldLoadResource(stateResource, configResource)) {
      this.loadResource(configResource);
    }
  }

  shouldLoadResource(stateResource: StateResource<T>, configResource: B): boolean {
    return isNotNull(configResource) && isLoadingRequired(stateResource);
  }

  loadResource(configResource: B): void {
    this.verifyGetLink(configResource);
    this.doLoadResource(getUrl(configResource, this.config.getLinkRel));
  }

  private verifyGetLink(configResource: B): void {
    throwErrorOn(
      !hasLink(configResource, this.config.getLinkRel),
      'No get link exists on configresource.',
    );
  }

  //TODO rename to reloadByResourceUri
  public setResourceByUri(resourceUri: ResourceUri): void {
    this.doLoadResource(resourceUri);
  }

  doLoadResource(resourceUri: ResourceUri): void {
    this.setStateResourceLoading();
    this.repository
      .getResource(resourceUri)
      .pipe(first())
      .subscribe((loadedResource: T) => this.updateStateResource(loadedResource));
  }

  setStateResourceLoading(): void {
    this.stateResource.next({ ...createEmptyStateResource(true), reload: false });
  }

  updateStateResource(resource: T): void {
    this.stateResource.next(createStateResource(resource));
  }

  handleNullConfigResource(configResource: B): void {
    if (this.shouldClearStateResource(configResource)) {
      this.stateResource.next(createEmptyStateResource());
    }
  }

  shouldClearStateResource(configResource: B): boolean {
    return isNull(configResource) && !isEqual(this.stateResource.value, createEmptyStateResource());
  }

  public save(toSave: unknown): Observable<StateResource<T | HttpError>> {
    this.verifyEditLinkRel();
    const previousResource: T = this.stateResource.value.resource;
    return this.doSave(previousResource, toSave).pipe(
      tap((loadedResource: T) => this.stateResource.next(createStateResource(loadedResource))),
      map(() => this.stateResource.value),
      catchError((errorResponse: HttpErrorResponse) => this.handleError(errorResponse)),
    );
  }

  handleError(errorResponse: HttpErrorResponse): Observable<StateResource<HttpError>> {
    if (isUnprocessableEntity(errorResponse.status)) {
      return of(createErrorStateResource((<any>errorResponse) as HttpError));
    }
    return throwError(() => errorResponse);
  }

  public verifyEditLinkRel(): void {
    throwErrorOn(
      !this.hasLinkRel(this.config.edit.linkRel),
      'No edit link exists on current stateresource.',
    );
  }

  abstract doSave(resource: T, toSave: unknown): Observable<T>;

  public refresh(): void {
    this.stateResource.next({ ...this.stateResource.value, reload: true });
  }

  public canEdit(): boolean {
    return this.hasLinkRel(this.config.edit.linkRel);
  }

  protected hasLinkRel(linkRel: string): boolean {
    return hasLink(this.getResource(), linkRel);
  }

  public getResource(): T {
    return this.stateResource.value.resource;
  }

  public canDelete(): boolean {
    return this.hasLinkRel(this.config.delete.linkRel);
  }

  protected verifyDeleteLinkRel(): void {
    throwErrorOn(!this.canDelete(), 'No delete link exists on current stateresource.');
  }

  public exists(): boolean {
    return isNotNull(this.stateResource.value.resource);
  }
}
