import {
  Bescheid,
  BescheidLinkRel,
  BescheidResource,
  BescheidSendBy,
  BescheidService,
  DocumentResource,
} from '@alfa-client/bescheid-shared';
import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import { CommandLinkRel, CommandResource } from '@alfa-client/command-shared';
import {
  EMPTY_STRING,
  StateResource,
  createStateResource,
  formatForDatabase,
} from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { VorgangWithEingangResource } from '@alfa-client/vorgang-shared';
import { registerLocaleData } from '@angular/common';
import localeDe from '@angular/common/locales/de';
import { fakeAsync, tick } from '@angular/core/testing';
import { UntypedFormBuilder } from '@angular/forms';
import faker from '@faker-js/faker';
import { Resource, ResourceUri } from '@ngxp/rest';
import { cold } from 'jest-marbles';
import { createCommandResource } from 'libs/command-shared/test/command';
import { toResource } from 'libs/tech-shared/test/resource';
import { of } from 'rxjs';
import {
  createBescheid,
  createBescheidResource,
} from '../../../../../bescheid-shared/src/test/bescheid';
import { createDocumentResource } from '../../../../../bescheid-shared/src/test/document';
import { createBinaryFileResource } from '../../../../../binary-file-shared/test/binary-file';
import { singleCold } from '../../../../../tech-shared/test/marbles';
import { createVorgangWithEingangResource } from '../../../../../vorgang-shared/test/vorgang';
import { BescheidenFormService } from './bescheiden.formservice';

registerLocaleData(localeDe);

describe('BescheidenFormService', () => {
  let service: BescheidenFormService;
  let bescheidService: Mock<BescheidService>;
  const now: Date = new Date();
  Date.now = jest.fn().mockReturnValue(now);

  beforeEach(() => {
    bescheidService = mock(BescheidService);
    service = new BescheidenFormService(new UntypedFormBuilder(), useFromMock(bescheidService));
  });

  describe('constructor', () => {
    it('should initialize form beschieden am', () => {
      expect(service.getValue().beschiedenAm).toEqual(formatForDatabase(now));
    });

    it('should initialize form bewilligt', () => {
      expect(service.getValue().bewilligt).toBeTruthy();
    });

    it('should initialize form send by', () => {
      expect(service.getValue().sendBy).toEqual(BescheidSendBy.NACHRICHT);
    });

    it('should initialize form bescheid dokument', () => {
      expect(service.getValue().bescheidDocument).toBeNull();
    });

    it('should initialize form attachments', () => {
      expect(service.getValue().attachments).toEqual([]);
    });

    it('should initialize form nachricht subject', () => {
      expect(service.getValue().nachrichtSubject).toEqual(EMPTY_STRING);
    });

    it('should initialize form nachricht text', () => {
      expect(service.getValue().nachrichtText).toEqual(EMPTY_STRING);
    });
  });

  describe('init', () => {
    it('should subscribe to sendBy', () => {
      service.subscribeToSendBy = jest.fn();

      service.init();

      expect(service.subscribeToSendBy).toHaveBeenCalled();
    });
  });

  describe('subscribe to send by', () => {
    it('should update send by on value change', () => {
      service.updateSendByManual = jest.fn();

      service.subscribeToSendBy();
      service.form.controls[BescheidenFormService.FIELD_SEND_BY].patchValue(BescheidSendBy.MANUAL);

      expect(service.updateSendByManual).toHaveBeenCalled();
    });
  });

  describe('update send by manual', () => {
    it('should set to true if its manual', () => {
      service.updateSendByManual(BescheidSendBy.MANUAL);

      expect(service.sendByManual.value).toBeTruthy();
    });
    it('should set to false if its nachricht', () => {
      service.updateSendByManual(BescheidSendBy.NACHRICHT);

      expect(service.sendByManual.value).toBeFalsy();
    });
  });

  describe('initializeFormChanges', () => {
    it('should emit initial form value', () => {
      expect(service.getBescheidChanges()).toBeObservable(
        cold('a', {
          a: {
            beschiedenAm: now,
            bewilligt: true,
            attachments: [],
            bescheidDocument: null,
            sendBy: BescheidSendBy.NACHRICHT,
            nachrichtSubject: '',
            nachrichtText: '',
          },
        }),
      );
    });
  });

  describe('getValue', () => {
    let getFormValue: jest.Mock;

    beforeEach(() => {
      getFormValue = service.getFormValue = jest.fn();
      getFormValue.mockReturnValue({ beschiedenAm: now, bewilligt: 'true' });
    });

    it('should call getFormValue', () => {
      service.getValue();

      expect(getFormValue).toHaveBeenCalled();
    });

    it('should return bescheid', () => {
      const value: Bescheid = service.getValue();

      expect(value).toEqual({ bewilligt: true, beschiedenAm: formatForDatabase(now) } as Bescheid);
    });
  });

  describe('getBescheidChanges', () => {
    it('should emit initial form value', () => {
      expect(service.getBescheidChanges()).toBeObservable(
        singleCold({
          beschiedenAm: now,
          bewilligt: true,
          attachments: [],
          bescheidDocument: null,
          sendBy: BescheidSendBy.NACHRICHT,
          nachrichtSubject: '',
          nachrichtText: '',
        }),
      );
    });
  });

  describe('patchValues', () => {
    let bescheidResource: BescheidResource;
    let patch: jest.Mock;

    beforeEach(() => {
      bescheidResource = createBescheidResource();
      patch = service.patch = jest.fn();
    });

    it('should call patch', () => {
      service.patchValues(bescheidResource);

      expect(patch).toHaveBeenCalledWith({
        [BescheidenFormService.FIELD_BESCHIEDEN_AM]: bescheidResource.beschiedenAm,
        [BescheidenFormService.FIELD_BEWILLIGT]: String(bescheidResource.bewilligt),
        [BescheidenFormService.FIELD_BESCHEID_DOCUMENT]: null,
        [BescheidenFormService.FIELD_SEND_BY]: String(bescheidResource.sendBy),
        [BescheidenFormService.FIELD_NACHRICHT_SUBJECT]: bescheidResource.nachrichtSubject,
        [BescheidenFormService.FIELD_NACHRICHT_TEXT]: bescheidResource.nachrichtText,
      });
    });

    it('should call patch with undefined sendBy', () => {
      const bescheidDocumentUri: ResourceUri = faker.internet.url();
      service.patchValues({
        ...bescheidResource,
        sendBy: undefined,
        _links: {
          ...bescheidResource._links,
          [BescheidLinkRel.BESCHEID_DOCUMENT]: { href: bescheidDocumentUri },
        },
      });

      expect(patch).toHaveBeenCalledWith({
        [BescheidenFormService.FIELD_BESCHIEDEN_AM]: bescheidResource.beschiedenAm,
        [BescheidenFormService.FIELD_BEWILLIGT]: String(bescheidResource.bewilligt),
        [BescheidenFormService.FIELD_SEND_BY]: BescheidSendBy.NACHRICHT,
        [BescheidenFormService.FIELD_BESCHEID_DOCUMENT]: bescheidDocumentUri,
        [BescheidenFormService.FIELD_NACHRICHT_SUBJECT]: bescheidResource.nachrichtSubject,
        [BescheidenFormService.FIELD_NACHRICHT_TEXT]: bescheidResource.nachrichtText,
      });
    });

    it('should not call patch', () => {
      service.patchValues(null);

      expect(patch).not.toHaveBeenCalled();
    });

    it('should patch attachments', () => {
      service.patchValues(bescheidResource);

      expect(service.getFormValue().attachments).toEqual(bescheidResource.attachments);
    });
  });

  describe('submit', () => {
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(createCommandResource());

    beforeEach(() => {
      bescheidService.updateBescheid.mockReturnValue(of(commandStateResource));
      bescheidService.createBescheid.mockReturnValue(of(commandStateResource));
    });

    it('should call bescheid service update bescheid draft if is patch', (done) => {
      const dummyPatchResource: Resource = toResource({});
      service.patch(dummyPatchResource);

      service.submit().subscribe(() => {
        expect(bescheidService.updateBescheid).toHaveBeenCalledWith(service.getValue());
        done();
      });
    });

    it('should call bescheid service update bescheid draft if bescheid draft exists', (done) => {
      bescheidService.existsBescheidDraft.mockReturnValue(true);

      service.submit().subscribe(() => {
        expect(bescheidService.updateBescheid).toHaveBeenCalledWith(service.getValue());
        done();
      });
    });

    it('should call bescheid service create bescheid draft if bescheid draft exists', (done) => {
      bescheidService.existsBescheidDraft.mockReturnValue(false);

      service.submit().subscribe(() => {
        expect(bescheidService.createBescheid).toHaveBeenCalledWith(
          service.vorgangWithEingangResource,
          service.getValue(),
        );
        done();
      });
    });
  });

  describe('submit draft', () => {
    const commandStateResource: StateResource<CommandResource> = createStateResource(
      createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]),
    );

    beforeEach(() => {
      service.submit = jest.fn().mockReturnValue(of(commandStateResource));
    });

    it('should call submit', (done) => {
      service.submitDraft().subscribe(() => {
        expect(service.submit).toHaveBeenCalledWith();
        done();
      });
    });

    it('should call service to reload vorgang', fakeAsync(() => {
      service.submitDraft().subscribe();
      tick();

      expect(bescheidService.reloadCurrentVorgang).toHaveBeenCalled();
    }));
  });

  describe('setVorgangWithEingangResource', () => {
    it('should set vorgangWithEingangResource', () => {
      const vorgangWithEingangResource: VorgangWithEingangResource =
        createVorgangWithEingangResource();

      service.setVorgangWithEingangResource(vorgangWithEingangResource);

      expect(service.vorgangWithEingangResource).toBe(vorgangWithEingangResource);
    });
  });

  describe('getVorgangWithEingangResource', () => {
    it('should reutrn vorgangWithEingangResource', () => {
      const vorgangWithEingangResource: VorgangWithEingangResource =
        createVorgangWithEingangResource();
      service.vorgangWithEingangResource = vorgangWithEingangResource;

      const vorang: VorgangWithEingangResource = service.getVorgangWithEingangResource();

      expect(vorang).toBe(vorgangWithEingangResource);
    });
  });

  describe('deleteFile', () => {
    it('should emit binary file resource', (done) => {
      const file: BinaryFileResource = createBinaryFileResource();

      service.getFileDelete().subscribe((deletedFile: BinaryFileResource) => {
        expect(deletedFile).toEqual(file);
        done();
      });

      service.deleteFile(file);
    });
  });

  describe('setActiveStep', () => {
    it('should emit step', () => {
      service.setActiveStep(3);

      expect(service.getActiveStep()).toBeObservable(singleCold(3));
    });
  });

  describe('patchNachricht', () => {
    it('should patch nachrichtSubject with value from document', () => {
      const documentResource: DocumentResource = createDocumentResource();

      service.patchNachricht(documentResource);

      expect(service.getValue().nachrichtSubject).toEqual(documentResource.nachrichtSubject);
    });

    it.each([EMPTY_STRING, null, undefined])(
      'should patch nachrichtSubject with default value if subject not defined',
      (subject: string) => {
        const documentResource: DocumentResource = {
          ...createDocumentResource(),
          nachrichtSubject: subject,
        };

        service.patchNachricht(documentResource);

        expect(service.getValue().nachrichtSubject).toEqual(BescheidenFormService.BETREFF_DEFAULT);
      },
    );

    it('should not patch nachrichtSubject if given', () => {
      const documentResource: DocumentResource = createDocumentResource();
      const bescheidResource: BescheidResource = createBescheidResource();
      service.patchValues(bescheidResource);

      service.patchNachricht(documentResource);

      expect(service.getValue().nachrichtSubject).toEqual(bescheidResource.nachrichtSubject);
    });

    it('should patch nachrichtText with value from document', () => {
      const documentResource: DocumentResource = createDocumentResource();

      service.patchNachricht(documentResource);

      expect(service.getValue().nachrichtText).toEqual(documentResource.nachrichtText);
    });

    it.each([EMPTY_STRING, null, undefined])(
      'should build Nachricht Text if text not defined',
      (text: string) => {
        service.buildNachrichtentext = jest.fn();
        const documentResource: DocumentResource = {
          ...createDocumentResource(),
          nachrichtText: text,
        };

        service.patchNachricht(documentResource);

        expect(service.buildNachrichtentext).toHaveBeenCalled();
      },
    );

    it.each([EMPTY_STRING, null, undefined])(
      'should patch nachrichtText with default value if text not defined',
      (text: string) => {
        const nachrichtText = faker.lorem.text();
        service.buildNachrichtentext = jest.fn().mockReturnValue(nachrichtText);
        const documentResource: DocumentResource = {
          ...createDocumentResource(),
          nachrichtText: text,
        };

        service.patchNachricht(documentResource);

        expect(service.getValue().nachrichtText).toEqual(nachrichtText);
      },
    );

    it('should not patch nachrichtText if given', () => {
      const documentResource: DocumentResource = createDocumentResource();
      const bescheidResource: BescheidResource = createBescheidResource();
      service.patchValues(bescheidResource);

      service.patchNachricht(documentResource);

      expect(service.getValue().nachrichtText).toEqual(bescheidResource.nachrichtText);
    });
  });

  describe('clearNachricht', () => {
    it('should patch values', () => {
      const documentResource: DocumentResource = createDocumentResource();

      service.patchNachricht(documentResource);
      service.clearNachricht();

      expect(service.getValue().nachrichtSubject).toEqual(EMPTY_STRING);
      expect(service.getValue().nachrichtText).toEqual(EMPTY_STRING);
    });
  });

  describe('setSendBy', () => {
    it.each(Object.keys(BescheidSendBy))(
      'should set control value to %s',
      (sendBy: BescheidSendBy) => {
        service.setSendBy(sendBy);

        expect(service.getValue().sendBy).toEqual(sendBy);
      },
    );
  });

  describe('buildNachrichtentext', () => {
    it('should build Nachrichtentext bewilligt', () => {
      service.getValue = jest
        .fn()
        .mockReturnValue({ ...createBescheid(), bewilligt: true } as Bescheid);

      const nachrichtentext = service.buildNachrichtentext();

      expect(nachrichtentext).toEqual(BescheidenFormService.NACHRICHTENTEXT_BEWILLIGT_DEFAULT);
    });

    it('should build Nachrichtentext abgelehnt', () => {
      service.getValue = jest
        .fn()
        .mockReturnValue({ ...createBescheid(), bewilligt: false } as Bescheid);

      const nachrichtentext = service.buildNachrichtentext();

      expect(nachrichtentext).toEqual(BescheidenFormService.NACHRICHTENTEXT_ABGELEHNT_DEFAULT);
    });
  });

  describe('validateBescheidDocumentExists', () => {
    it('should return true', () => {
      service.getValue = jest.fn().mockReturnValue(createBescheid());

      service.validateBescheidDocumentExists();

      expect(service.validateBescheidDocumentExists()).toBeTruthy();
    });

    it('should return false', () => {
      service.getValue = jest.fn().mockReturnValue({
        ...createBescheid(),
        bescheidDocument: null,
      } as Bescheid);

      service.validateBescheidDocumentExists();

      expect(service.validateBescheidDocumentExists()).toBeFalsy();
    });

    describe('getShowMissingBescheidDocumentError', () => {
      it('should emit true', () => {
        service.getValue = jest.fn().mockReturnValue({
          ...createBescheid(),
          bescheidDocument: null,
        } as Bescheid);

        service.validateBescheidDocumentExists();

        expect(service.getShowMissingBescheidDocumentError()).toBeObservable(singleCold(true));
      });

      it('should emit false', () => {
        service.getValue = jest.fn().mockReturnValue(createBescheid());

        service.validateBescheidDocumentExists();

        expect(service.getShowMissingBescheidDocumentError()).toBeObservable(singleCold(false));
      });
    });
  });
});
