import { BescheidService } from '@alfa-client/bescheid-shared';
import { BinaryFile2ContainerComponent } from '@alfa-client/binary-file';
import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import {
  convertForDataTest,
  ConvertForDataTestPipe,
  createEmptyStateResource,
  createErrorStateResource,
  createStateResource,
  FileSizePipe,
  StateResource,
} from '@alfa-client/tech-shared';
import { existsAsHtmlElement, Mock, mock, notExistsAsHtmlElement } from '@alfa-client/test-utils';
import { OzgcloudSvgIconComponent, SpinnerComponent } from '@alfa-client/ui';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { MatIcon } from '@angular/material/icon';
import {
  AttachmentComponent,
  AttachmentContainerComponent,
  SpinnerIconComponent,
} from '@ods/system';
import { MockComponent, MockPipe } from 'ng-mocks';
import { BehaviorSubject, EMPTY, Observable, of, Subscription } from 'rxjs';
import {
  createBinaryFileResource,
  createLoadedBinaryFileResource,
  createLoadingBinaryFileStateResource,
} from '../../../../../../../binary-file-shared/test/binary-file';
import { getDataTestIdOf } from '../../../../../../../tech-shared/test/data-test';
import { createApiError } from '../../../../../../../tech-shared/test/error';
import { BescheidenFormService } from '../../bescheiden.formservice';
import { VorgangDetailBescheidenFormErrorComponent } from '../vorgang-detail-bescheiden-form-error/vorgang-detail-bescheiden-form-error.component';
import { VorgangDetailBescheidenResultAttachmentsComponent } from './vorgang-detail-bescheiden-result-attachments.component';

describe('VorgangDetailBescheidenResultAttachmentsComponent', () => {
  let component: VorgangDetailBescheidenResultAttachmentsComponent;
  let fixture: ComponentFixture<VorgangDetailBescheidenResultAttachmentsComponent>;

  let bescheidService: Mock<BescheidService>;
  let formService: Mock<BescheidenFormService>;

  beforeEach(async () => {
    bescheidService = mock(BescheidService);
    bescheidService.getAttachments.mockReturnValue(EMPTY);

    formService = mock(BescheidenFormService);
    formService.getBescheidChanges.mockReturnValue(
      new BehaviorSubject({ beschiedenAm: new Date(), bewilligt: false }),
    );
    formService.getAttachmentUpload.mockReturnValue(of(createEmptyStateResource()));
    formService.getBescheidFileUpload.mockReturnValue(of(createEmptyStateResource()));

    await TestBed.configureTestingModule({
      declarations: [
        VorgangDetailBescheidenResultAttachmentsComponent,
        ConvertForDataTestPipe,
        MatIcon,
        MockPipe(FileSizePipe),
        MockComponent(OzgcloudSvgIconComponent),
        MockComponent(SpinnerComponent),
        MockComponent(VorgangDetailBescheidenFormErrorComponent),
        MockComponent(AttachmentContainerComponent),
        MockComponent(BinaryFile2ContainerComponent),
        MockComponent(SpinnerIconComponent),
        MockComponent(AttachmentComponent),
      ],
      providers: [
        {
          provide: BescheidenFormService,
          useValue: formService,
        },
        {
          provide: BescheidService,
          useValue: bescheidService,
        },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(VorgangDetailBescheidenResultAttachmentsComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('render', () => {
    const attachment = createStateResource(createBinaryFileResource());
    const dataTestId = getDataTestIdOf(
      `${convertForDataTest(attachment.resource.name)}-file2-container`,
    );

    beforeEach(() => {
      component.uploadedAttachments = [attachment];
    });

    it('should render uploaded file container', () => {
      component.ngOnInit();
      fixture.detectChanges();

      existsAsHtmlElement(fixture, dataTestId);
    });

    it('should not render uploaded file container on loading', () => {
      component.uploadedAttachments = [{ ...attachment, loading: true }];
      component.ngOnInit();
      fixture.detectChanges();

      notExistsAsHtmlElement(fixture, dataTestId);
    });

    it('should not render uploaded file container on null resource', () => {
      component.uploadedAttachments = [{ ...attachment, resource: null }];
      component.ngOnInit();
      fixture.detectChanges();

      notExistsAsHtmlElement(fixture, dataTestId);
    });
  });

  describe('ngOnInit', () => {
    let loadExistingAttachments: jest.Mock;
    let subscribeToAttachmentUpload: jest.Mock;

    beforeEach(() => {
      component.loadExistingAttachments = loadExistingAttachments = jest.fn();
      component.subscribeToAttachmentUpload = subscribeToAttachmentUpload = jest.fn();
    });

    it('should load existing attachments', () => {
      component.ngOnInit();

      expect(loadExistingAttachments).toHaveBeenCalled();
    });

    it('should subscribe to attachments upload', () => {
      component.ngOnInit();

      expect(subscribeToAttachmentUpload).toHaveBeenCalled();
    });
  });

  describe('ngOnDestroy', () => {
    it('should unsubscribe', () => {
      const subscription = mock(Subscription);
      const observable = mock(Observable);
      observable.subscribe.mockReturnValue(subscription);
      formService.getAttachmentUpload.mockReturnValue(observable);
      component.ngOnInit();

      component.ngOnDestroy();

      expect(subscription.unsubscribe).toHaveBeenCalled();
    });
  });

  describe('loadExistingAttachments', () => {
    let attachment: BinaryFileResource;

    beforeEach(() => {
      attachment = createBinaryFileResource();
      bescheidService.getAttachments.mockReturnValue(of([attachment]));
    });

    it('should get attachments', () => {
      component.loadExistingAttachments();

      expect(bescheidService.getAttachments).toHaveBeenCalled();
    });

    it('should set existing attachments', () => {
      component.loadExistingAttachments();

      expect(component.existingAttachments).toEqual([attachment]);
    });

    it('should reset uploaded attachments', () => {
      component.uploadedAttachments = [createStateResource(createBinaryFileResource())];

      component.loadExistingAttachments();

      expect(component.uploadedAttachments).toEqual([]);
    });
  });

  describe('subscribeToAttachmentUpload', () => {
    let attachment: StateResource<BinaryFileResource>;

    beforeEach(() => {
      attachment = createLoadedBinaryFileResource();
      formService.getAttachmentUpload.mockReturnValue(of(attachment));
    });

    it('should get attachment upload', () => {
      component.subscribeToAttachmentUpload();

      expect(formService.getAttachmentUpload).toHaveBeenCalled();
    });

    it('should should set uploaded attachments', () => {
      component.buildUploadedAttachments = jest.fn().mockReturnValue([attachment]);

      component.subscribeToAttachmentUpload();

      expect(component.uploadedAttachments).toEqual([attachment]);
    });
  });

  describe('buildUploadedAttachments', () => {
    let uploadedAttachment: StateResource<BinaryFileResource>;
    let uploadedAttachments: StateResource<BinaryFileResource>[];
    let loadingAttachment: StateResource<BinaryFileResource>;

    beforeEach(() => {
      uploadedAttachment = createLoadedBinaryFileResource();
      uploadedAttachments = [uploadedAttachment];
      component.uploadedAttachments = uploadedAttachments;
      loadingAttachment = createLoadingBinaryFileStateResource();
    });

    it('should add loaded state resource', () => {
      const newBinaryFileResource = createLoadedBinaryFileResource();

      const uploaded = component.buildUploadedAttachments(newBinaryFileResource);

      expect(uploaded).toEqual([uploadedAttachment, newBinaryFileResource]);
    });

    it('should add error state resource', () => {
      const errorStateResource = createErrorStateResource(createApiError());

      const uploaded = component.buildUploadedAttachments(errorStateResource);

      expect(uploaded).toEqual([uploadedAttachment, errorStateResource]);
    });

    it('should add loading state resource if none exists', () => {
      const uploaded = component.buildUploadedAttachments(loadingAttachment);

      expect(uploaded).toEqual([uploadedAttachment, loadingAttachment]);
    });

    it('should not add loading state resource if one exists', () => {
      component.uploadedAttachments = [...component.uploadedAttachments, loadingAttachment];

      const uploaded = component.buildUploadedAttachments(createLoadingBinaryFileStateResource());

      expect(uploaded).toEqual([uploadedAttachment, loadingAttachment]);
    });

    it('should should not have error state resource if loading in progress', () => {
      component.uploadedAttachments = [createErrorStateResource(createApiError())];

      const uploaded = component.buildUploadedAttachments(loadingAttachment);

      expect(uploaded).toEqual([loadingAttachment]);
    });
  });

  describe('deleteFile', () => {
    let existingAttachment: BinaryFileResource;
    let uploadedAttachmentStateResource: StateResource<BinaryFileResource>;

    beforeEach(() => {
      existingAttachment = createBinaryFileResource();
      uploadedAttachmentStateResource = createStateResource(createBinaryFileResource());
    });

    it('should call form service', () => {
      const file = createBinaryFileResource();

      component.deleteFile(file);

      expect(formService.deleteFile).toHaveBeenCalledWith(file);
    });

    it('should delete attachment from uploaded', () => {
      component.uploadedAttachments = [uploadedAttachmentStateResource];

      component.deleteFile(uploadedAttachmentStateResource.resource);

      expect(component.uploadedAttachments).toEqual([]);
    });

    it('should not delete attachment from existing', () => {
      component.uploadedAttachments = [uploadedAttachmentStateResource];
      component.existingAttachments = [existingAttachment];

      component.deleteFile(uploadedAttachmentStateResource.resource);

      expect(component.existingAttachments).toEqual([existingAttachment]);
    });

    it('should delete attachment from existing', () => {
      component.existingAttachments = [existingAttachment];

      component.deleteFile(existingAttachment);

      expect(component.existingAttachments).toEqual([]);
    });

    it('should not delete attachment from existing', () => {
      component.uploadedAttachments = [uploadedAttachmentStateResource];
      component.existingAttachments = [existingAttachment];

      component.deleteFile(existingAttachment);

      expect(component.uploadedAttachments).toEqual([uploadedAttachmentStateResource]);
    });
  });
});
