import {
  Bescheid,
  BescheidLinkRel,
  BescheidResource,
  BescheidService,
  DocumentResource,
  UploadFileInProgress,
} from '@alfa-client/bescheid-shared';
import { BinaryFileResource } from '@alfa-client/binary-file-shared';
import {
  CommandOrder,
  CommandResource,
  switchMapCommandSuccessfullyDone,
  tapOnCommandSuccessfullyDone,
} from '@alfa-client/command-shared';
import { StateResource, createEmptyStateResource, isLoaded } from '@alfa-client/tech-shared';
import { Component, EventEmitter, OnInit, Output } from '@angular/core';
import { hasLink } from '@ngxp/rest';
import { Observable, filter, map, of } from 'rxjs';
import { BescheidenFormService } from '../bescheiden.formservice';

type sendBescheid = (
  BescheidResource: BescheidResource,
) => Observable<StateResource<CommandResource>>;

@Component({
  selector: 'alfa-vorgang-detail-bescheiden-result',
  templateUrl: './vorgang-detail-bescheiden-result.component.html',
})
export class VorgangDetailBescheidenResultComponent implements OnInit {
  @Output() closeDialog: EventEmitter<void> = new EventEmitter();

  public activeStep$: Observable<number>;
  public bescheid$: Observable<Bescheid>;
  public sendByManual$: Observable<boolean>;

  public bescheidDraftStateResource$: Observable<StateResource<BescheidResource>> = of(
    createEmptyStateResource<BescheidResource>(),
  );
  public bescheidDocumentFile$: Observable<StateResource<BinaryFileResource>> = of(
    createEmptyStateResource<BinaryFileResource>(),
  );
  public bescheidDocument$: Observable<StateResource<DocumentResource>> = of(
    createEmptyStateResource<DocumentResource>(),
  );
  public saveAndSendInProgress$: Observable<StateResource<CommandResource>> = of(
    createEmptyStateResource<CommandResource>(),
  );

  public uploadBescheidDocumentInProgress$: Observable<UploadFileInProgress> = of({
    loading: false,
  });

  public createBescheidDocumentInProgress$: Observable<UploadFileInProgress> = of({
    loading: false,
  });

  public canSave$: Observable<boolean> = of(true);
  public canSend$: Observable<boolean> = of(true);
  public showMissingBescheidDocumentError$: Observable<boolean> = of(false);

  public bescheidLinkRel = BescheidLinkRel;

  constructor(
    private bescheidService: BescheidService,
    public formService: BescheidenFormService,
  ) {}

  ngOnInit(): void {
    this.bescheidDraftStateResource$ = this.bescheidService.getBescheidDraft();
    this.bescheidDocumentFile$ = this.bescheidService.getBescheidDocumentFile();
    this.uploadBescheidDocumentInProgress$ =
      this.bescheidService.getUploadBescheidDocumentInProgress();
    this.createBescheidDocumentInProgress$ =
      this.bescheidService.getCreateBescheidDocumentInProgress();
    this.bescheidDocument$ = this.bescheidService.getBescheidDocument();

    this.activeStep$ = this.formService.getActiveStep();
    this.bescheid$ = this.formService.getBescheidChanges();
    this.sendByManual$ = this.formService.isSendByManual();

    this.canSave$ = this.bescheidDraftStateResource$.pipe(
      filter(isLoaded),
      map((stateResource: StateResource<BescheidResource>) =>
        hasLink(stateResource.resource, BescheidLinkRel.BESCHEIDEN),
      ),
    );
    this.canSend$ = this.bescheidDraftStateResource$.pipe(
      filter(isLoaded),
      map((stateResource: StateResource<BescheidResource>) =>
        hasLink(stateResource.resource, BescheidLinkRel.BESCHEIDEN_UND_SENDEN),
      ),
    );
    this.showMissingBescheidDocumentError$ = this.formService.getShowMissingBescheidDocumentError();
  }

  public deleteBescheidDocument(): void {
    this.formService.clearBescheidDocumentFile();
    this.bescheidService.deleteBescheidDocument();
  }

  public saveAndSendManually(bescheidDraft: BescheidResource): void {
    this.formService.clearNachricht();
    this.saveAndSendInProgress$ = this.doUpdateAndSend(bescheidDraft, (bescheidResource) =>
      this.bescheidService.sendBescheidManually(bescheidResource),
    );
  }

  public saveAndSendWithNachricht(bescheidDraft: BescheidResource): void {
    this.saveAndSendInProgress$ = this.doUpdateAndSend(bescheidDraft, (bescheidResource) =>
      this.bescheidService.sendBescheidToAntragsteller(bescheidResource),
    );
  }

  doUpdateAndSend(
    bescheidDraft: BescheidResource,
    send: sendBescheid,
  ): Observable<StateResource<CommandResource>> {
    return this.formService.submit().pipe(
      switchMapCommandSuccessfullyDone(() => send(bescheidDraft)),
      tapOnCommandSuccessfullyDone((commandStateResource: StateResource<CommandResource>) => {
        if (this.isSendBescheidCommand(commandStateResource)) {
          this.closeDialog.emit();
        }
      }),
    );
  }

  private isSendBescheidCommand(commandStateResource: StateResource<CommandResource>): boolean {
    return commandStateResource.resource.order == CommandOrder.SEND_BESCHEID;
  }
}
