package de.ozgcloud.alfa.bescheid;

import static de.ozgcloud.alfa.common.UserProfileUrlProviderTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.function.BooleanSupplier;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;

import de.ozgcloud.alfa.common.FeatureToggleProperties;
import de.ozgcloud.alfa.common.UserProfileUrlProvider;
import de.ozgcloud.alfa.vorgang.Vorgang;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import lombok.NoArgsConstructor;

class BescheidVorgangProcessorTest {

	private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

	@Spy
	@InjectMocks
	private BescheidVorgangProcessor processor;
	@Mock
	private BescheidService service;

	@Mock
	private FeatureToggleProperties featureToggleProperties;

	@Nested
	class TestProcess {

		@Test
		void shouldReturnTheSameModelOnNullVorgang() {
			var inputModel = new NullableEntityModel();

			var processedModel = processor.process(inputModel);

			assertThat(processedModel).isEqualTo(inputModel);
		}

		@Test
		void shouldCallExistsBescheid() {
			initUserProfileUrlProvider(new UserProfileUrlProvider());
			doReturn(true).when(processor).existsBescheid(anyString());

			callProcess();

			verify(processor).existsBescheid(VorgangHeaderTestFactory.ID);
		}

		@Nested
		class TestLinks {

			@BeforeEach
			void init() {
				initUserProfileUrlProvider(new UserProfileUrlProvider());
			}

			@Test
			void shouldNotHaveLinkToDraft() {
				givenRetrievingDraftIsAllowed(false);

				var model = callProcess();

				assertThat(model.getLink(BescheidVorgangProcessor.REL_DRAFT)).isEmpty();
			}

			@Test
			void shouldHaveLinkToDraft() {
				givenRetrievingDraftIsAllowed(true);

				var model = callProcess();

				assertThat(model.getLink(BescheidVorgangProcessor.REL_DRAFT)).isPresent().get().extracting(Link::getHref)
						.isEqualTo(BescheidController.PATH + "?vorgangId=" + VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveLinkToCreateBescheidDraft() {
				givenCreatingDraftIsAllowed(true);

				var model = callProcess();

				assertThat(model.getLink(BescheidVorgangProcessor.REL_CREATE_DRAFT)).isPresent().get()
						.extracting(Link::getHref)
						.isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/relations/" + VorgangHeaderTestFactory.ID + "/"
								+ VorgangHeaderTestFactory.VERSION + "/commands");
			}

			@Test
			void shouldNotHaveLinkToCreateBescheidDraft() {
				givenCreatingDraftIsAllowed(false);

				var model = callProcess();

				assertThat(model.getLink(BescheidVorgangProcessor.REL_CREATE_DRAFT)).isEmpty();
			}

			@DisplayName("bescheide link")
			@Nested
			class TestBescheideLink {

				@DisplayName("on existing bescheid")
				@Nested
				class TestOnExistingBescheid {

					@BeforeEach
					void mock() {
						doReturn(true).when(processor).existsBescheid(anyString());
					}

					@Test
					void shouldExists() {
						var model = callProcess();

						assertThat(model.getLink(BescheidVorgangProcessor.REL_BESCHEIDE)).isPresent().get()
								.extracting(Link::getHref)
								.isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/bescheids");
					}
				}

				@DisplayName("on missing bescheid")
				@Nested
				class TestOnMissingBescheid {

					@BeforeEach
					void mock() {
						doReturn(false).when(processor).existsBescheid(anyString());
					}

					@Test
					void shouldNOTExists() {
						var model = callProcess();

						assertThat(model.getLink(BescheidVorgangProcessor.REL_BESCHEIDE)).isNotPresent();
					}
				}
			}

			private void givenRetrievingDraftIsAllowed(boolean shouldBeAdded) {
				doReturn((BooleanSupplier) () -> shouldBeAdded).when(processor).isRetrievingDraftAllowed(vorgang);
			}

			private void givenCreatingDraftIsAllowed(boolean shouldBeAdded) {
				doReturn((BooleanSupplier) () -> shouldBeAdded).when(processor).isCreatingDraftAllowed(vorgang);
			}
		}

		private EntityModel<VorgangWithEingang> callProcess() {
			return processor.process(EntityModel.of(vorgang));
		}

		@NoArgsConstructor
		private static class NullableEntityModel extends EntityModel<VorgangWithEingang> {

		}
	}

	@DisplayName("Exists bescheid")
	@Nested
	class TestExistsBescheid {

		@BeforeEach
		void mock() {
			when(service.existsBescheid(anyString())).thenReturn(true);
		}

		@Test
		void shouldCallService() {
			processor.existsBescheid(VorgangHeaderTestFactory.ID);

			verify(service).existsBescheid(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var exists = processor.existsBescheid(VorgangHeaderTestFactory.ID);

			assertThat(exists).isTrue();
		}
	}

	@Nested
	class TestIsRetrievingDraftAllowed {

		@Test
		void shouldReturnTrue() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(true);
			givenDraftExists(true);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isTrue();
		}

		@Test
		void shouldReturnFalseIfFeatureToggleIsDisabled() {
			givenFeatureToggleEnabled(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		@Test
		void shouldReturnFalseIfVorgangNotInBearbeitung() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		@Test
		void shouldReturnFalseIfDraftDoesNotExist() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(true);
			givenDraftExists(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		private BooleanSupplier callMethod() {
			return processor.isRetrievingDraftAllowed(vorgang);
		}
	}

	@Nested
	class TestIsCreatingDraftAllowed {

		@Test
		void shouldReturnTrue() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(true);
			givenDraftExists(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isTrue();
		}

		@Test
		void shouldReturnFalseIfFeatureToggleIsDisabled() {
			givenFeatureToggleEnabled(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		@Test
		void shouldReturnFalseIfVorgangNotInBearbeitung() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(false);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		@Test
		void shouldReturnFalseIfDraftExists() {
			givenFeatureToggleEnabled(true);
			givenVorgangInBearbeitung(true);
			givenDraftExists(true);

			var booleanSupplier = callMethod();

			assertThat(booleanSupplier.getAsBoolean()).isFalse();
		}

		private BooleanSupplier callMethod() {
			return processor.isCreatingDraftAllowed(vorgang);
		}
	}

	@Nested
	class TestIsVorgangInBearbeitung {

		@ParameterizedTest
		@EnumSource(mode = EnumSource.Mode.EXCLUDE, names = { "IN_BEARBEITUNG" })
		void shouldReturnFalse(Vorgang.VorgangStatus vorgangStatus) {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(vorgangStatus).build();

			var inBearbeitung = processor.isVorgangInBearbeitung(vorgang);

			assertThat(inBearbeitung).isFalse();
		}

		@Test
		void shouldReturnTrue() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(Vorgang.VorgangStatus.IN_BEARBEITUNG).build();

			var inBearbeitung = processor.isVorgangInBearbeitung(vorgang);

			assertThat(inBearbeitung).isTrue();
		}
	}

	@Nested
	class TestDraftExists {

		@Test
		void shouldCallBescheidService() {
			processor.draftExists(vorgang);

			verify(service).getBescheidDraft(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldReturnFalse() {
			when(service.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(Optional.empty());

			var result = processor.draftExists(vorgang);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnTrue() {
			when(service.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(Optional.of(BescheidTestFactory.create()));

			var result = processor.draftExists(vorgang);

			assertThat(result).isTrue();
		}
	}

	private void givenFeatureToggleEnabled(boolean enabled) {
		when(featureToggleProperties.isBescheidWizard()).thenReturn(enabled);
	}

	private void givenVorgangInBearbeitung(boolean inBearbeitung) {
		doReturn(inBearbeitung).when(processor).isVorgangInBearbeitung(vorgang);
	}

	private void givenDraftExists(boolean exists) {
		doReturn(exists).when(processor).draftExists(vorgang);
	}
}
