package de.ozgcloud.alfa.historie;

import static de.ozgcloud.alfa.common.TestUtils.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.LocalDateTime;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserProfile;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserService;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.alfa.vorgang.ZustaendigeStelleTestFactory;

public class VorgangChangeHistoryServiceTest {

	private static final String ORGANISATIONSEINHEITEN_ID = ZustaendigeStelleTestFactory.ORGANISATIONSEINHEITEN_ID;

	private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
	private final Command command0105 = ChangeHistoryBuilderTest.commandFinishedAt(LocalDateTime.of(2023, 5, 1, 12, 0));
	private final Command command0106 = ChangeHistoryBuilderTest.commandFinishedAt(LocalDateTime.of(2023, 6, 1, 12, 0));
	private final List<Command> commands = List.of(command0105, command0106);

	private final VorgangChange vorgangChange0105 = VorgangChangeTestFactory.createBuilder()
			.valueBeforeChange(LoremIpsum.getInstance().getWords(2))
			.valueAfterChange(LoremIpsum.getInstance().getWords(2))
			.build();
	private final VorgangChange vorgangChange0106 = VorgangChangeTestFactory.createBuilder()
			.valueBeforeChange(LoremIpsum.getInstance().getWords(2))
			.valueAfterChange(LoremIpsum.getInstance().getWords(2))
			.build();
	private final List<VorgangChange> vorgangChanges = List.of(vorgangChange0105, vorgangChange0106);

	@Mock
	private UserService userService;
	@Mock
	private HistorieService historieService;
	@Mock
	private CommandService commandService;

	@Spy
	@InjectMocks
	private VorgangChangeHistoryService service;

	@Nested
	class TestCreateVorgangChangeHistory {

		@BeforeEach
		void init() {
			mockStreamToList(commands, stream -> when(commandService.findFinishedCommands(vorgang.getId())).thenReturn(stream));
		}

		@Test
		void shouldFindFinishedCommands() {
			callService();

			verify(commandService).findFinishedCommands(vorgang.getId());
		}

		@Test
		void shouldSetStatusChangeHistory() {
			doReturn(vorgangChanges.stream()).when(service).createStatusChangeHistory(vorgang, commands);

			var history = callService();

			assertThat(history.getStatusChangeHistory()).isEqualTo(vorgangChanges);
		}

		@Test
		void shouldSetAktenzeichenChangeHistory() {
			doReturn(vorgangChanges.stream()).when(service).createAktenzeichenChangeHistory(vorgang, commands);

			var history = callService();

			assertThat(history.getAktenzeichenChangeHistory()).isEqualTo(vorgangChanges);
		}

		@Test
		void shouldSetAssignedUserChangeHistory() {
			doReturn(vorgangChanges.stream()).when(service).createAssignedUserChangeHistory(vorgang, commands);

			var history = callService();

			assertThat(history.getAssignedUserChangeHistory()).isEqualTo(vorgangChanges);
		}

		private VorgangChangeHistory callService() {
			return service.createVorgangChangeHistory(vorgang);
		}
	}

	@Nested
	class TestCreateStatusChangeHistory {

		private MockedStatic<StatusChangeHistoryBuilder> staticBuilder;
		@Mock
		private StatusChangeHistoryBuilder builder;

		@BeforeEach
		void init() {
			mockBuilderFactoryMethod();
			mockBuilderWiths();
		}

		private void mockBuilderFactoryMethod() {
			staticBuilder = mockStatic(StatusChangeHistoryBuilder.class);
			staticBuilder.when(StatusChangeHistoryBuilder::builder).thenReturn(builder);
		}

		private void mockBuilderWiths() {
			when(builder.withCommands(commands)).thenReturn(builder);
			when(builder.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID)).thenReturn(builder);
		}

		@AfterEach
		void cleanup() {
			staticBuilder.close();
		}

		@Test
		void shouldSetCommands() {
			callService();

			verify(builder).withCommands(commands);
		}

		@Test
		void shouldSetOrganisationseinheitenID() {
			callService();

			verify(builder).withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);
		}

		@Test
		void shouldReturnBuiltStream() {
			when(builder.build()).thenReturn(vorgangChanges.stream());

			var changeHistory = callService();

			assertThat(changeHistory).containsExactlyElementsOf(vorgangChanges);
		}

		private Stream<VorgangChange> callService() {
			return service.createStatusChangeHistory(vorgang, commands);
		}
	}

	@Nested
	class TestCreateAktenzeichenChangeHistory {

		private MockedStatic<AktenzeichenChangeHistoryBuilder> staticBuilder;
		@Mock
		private AktenzeichenChangeHistoryBuilder builder;

		@BeforeEach
		void init() {
			mockBuilderFactoryMethod();
			mockBuilderWiths();
		}

		private void mockBuilderFactoryMethod() {
			staticBuilder = mockStatic(AktenzeichenChangeHistoryBuilder.class);
			staticBuilder.when(AktenzeichenChangeHistoryBuilder::builder).thenReturn(builder);
		}

		private void mockBuilderWiths() {
			when(builder.withCommands(commands)).thenReturn(builder);
			when(builder.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID)).thenReturn(builder);
		}

		@AfterEach
		void cleanup() {
			staticBuilder.close();
		}

		@Test
		void shouldSetCommands() {
			callService();

			verify(builder).withCommands(commands);
		}

		@Test
		void shouldSetOrganisationseinheitenID() {
			callService();

			verify(builder).withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);
		}

		@Test
		void shouldReturnBuiltStream() {
			when(builder.build()).thenReturn(vorgangChanges.stream());

			var changeHistory = callService();

			assertThat(changeHistory).containsExactlyElementsOf(vorgangChanges);
		}

		private Stream<VorgangChange> callService() {
			return service.createAktenzeichenChangeHistory(vorgang, commands);
		}
	}

	@Nested
	class TestCreateAssignedUserChangeHistory {

		private MockedStatic<AssignedUserChangeHistoryBuilder> staticBuilder;
		private MockedStatic<UserProfileCache> staticUserProfileCache;

		@Captor
		private ArgumentCaptor<Function<UserId, UserProfile>> userProfileCacheArgCaptor;

		@Mock
		private AssignedUserChangeHistoryBuilder builder;
		@Mock
		private UserProfileCache userProfileCache;

		@BeforeEach
		void init() {
			mockBuilderFactoryMethod();
			mockBuilderWiths();
			mockUserProfileCacheFactoryMethod();
		}

		private void mockBuilderFactoryMethod() {
			staticBuilder = mockStatic(AssignedUserChangeHistoryBuilder.class);
			staticBuilder.when(AssignedUserChangeHistoryBuilder::builder).thenReturn(builder);
		}

		private void mockBuilderWiths() {
			when(builder.withCommands(commands)).thenReturn(builder);
			when(builder.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID)).thenReturn(builder);
			when(builder.withUserProfileCache(userProfileCache)).thenReturn(builder);
		}

		private void mockUserProfileCacheFactoryMethod() {
			staticUserProfileCache = mockStatic(UserProfileCache.class);
			staticUserProfileCache.when(() -> UserProfileCache.create(any())).thenReturn(userProfileCache);
		}

		@AfterEach
		void cleanup() {
			staticBuilder.close();
			staticUserProfileCache.close();
		}

		@Test
		void shouldSetCommands() {
			callService();

			verify(builder).withCommands(commands);
		}

		@Test
		void shouldSetOrganisationseinheitenID() {
			callService();

			verify(builder).withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);
		}

		@Test
		void shouldCreateUserProfileCache() {
			callService();

			staticUserProfileCache.verify(() -> UserProfileCache.create(userProfileCacheArgCaptor.capture()));
			assertArgumentToUserProfileCacheCallsUserService();
		}

		private void assertArgumentToUserProfileCacheCallsUserService() {
			userProfileCacheArgCaptor.getValue().apply(UserProfileTestFactory.ID);

			verify(userService).getById(UserProfileTestFactory.ID);
		}

		@Test
		void shouldSetUserProfileCache() {
			callService();

			verify(builder).withUserProfileCache(userProfileCache);
		}

		@Test
		void shouldReturnBuiltStream() {
			when(builder.build()).thenReturn(vorgangChanges.stream());

			var changeHistory = callService();

			assertThat(changeHistory).containsExactlyElementsOf(vorgangChanges);
		}

		private Stream<VorgangChange> callService() {
			return service.createAssignedUserChangeHistory(vorgang, commands);
		}
	}

}
